"""
Functional tests for configuration of a build
"""
import json
import logging
import subprocess

import pytest

from ansible_bender.builders.buildah_builder import buildah

from tests.spellbook import b_p_w_vars_path, p_w_vars_files_path, data_dir, basic_playbook_path
from ..conftest import ab


logger = logging.getLogger("ansible_bender")


def test_basic(tmpdir):
    cmd = ["build", b_p_w_vars_path]
    ab(cmd, str(tmpdir))

    try:
        cmd = ["inspect", "--json"]
        ab_inspect_data = json.loads(ab(cmd, str(tmpdir), return_output=True))

        assert ab_inspect_data["base_image"] == "docker.io/library/python:3-alpine"
        assert ab_inspect_data["build_id"] == "1"
        assert ab_inspect_data['build_volumes'] == [f'{data_dir}:/src:Z']
        assert ab_inspect_data['builder_name'] == "buildah"
        assert len(ab_inspect_data['layers']) == 3
        assert ab_inspect_data["metadata"]["labels"] == {"x": "y"}
        assert ab_inspect_data["metadata"]["env_vars"] == {"asd": data_dir}
        assert ab_inspect_data["playbook_path"] == b_p_w_vars_path
        assert ab_inspect_data["pulled"] is False
        assert ab_inspect_data["target_image"] == "challet"

        cmd = ["podman", "inspect", "--type", "image", "challet"]
        inspect_data = json.loads(subprocess.check_output(cmd))[0]

        assert inspect_data["Config"]["Labels"] == {"x": "y"}
        assert f"asd={data_dir}" in inspect_data["Config"]["Env"]
        assert inspect_data["Config"]["WorkingDir"] == "/src"
    finally:
        try:
            buildah("rmi", ["challet"])  # FIXME: use builder interface instead for sake of other backends
        except subprocess.CalledProcessError as ex:
            print(ex)


def test_with_vars_files(tmpdir):
    cmd = ["build", p_w_vars_files_path]
    ab(cmd, str(tmpdir))

    try:
        cmd = ["inspect", "--json"]
        ab_inspect_data = json.loads(ab(cmd, str(tmpdir), return_output=True))

        assert ab_inspect_data["base_image"] == "docker.io/library/python:3-alpine"
        assert ab_inspect_data["build_id"] == "1"
        assert ab_inspect_data['builder_name'] == "buildah"
        assert len(ab_inspect_data['layers']) == 3
        assert ab_inspect_data["metadata"]["labels"] == {"x": "y"}
        assert ab_inspect_data["metadata"]["env_vars"] == {"key": "env", "path": "/etc/passwd"}
        assert ab_inspect_data["playbook_path"] == p_w_vars_files_path
        assert ab_inspect_data["pulled"] is False
        assert ab_inspect_data["target_image"] == "with-vars-files"
    finally:
        try:
            buildah("rmi", ["with-vars-files"])  # FIXME: use builder interface instead for sake of other backends
        except subprocess.CalledProcessError as ex:
            print(ex)


def test_basic_build_errr(tmpdir):
    cmd = ["build", basic_playbook_path]
    with pytest.raises(subprocess.CalledProcessError) as ex:
        ab(cmd, str(tmpdir), return_output=True)
    e = ex.value.stdout
    assert "Failed validating 'type' in schema['properties']['base_image']:" in e
    assert "There was an error during execution: None is not of type 'string'" in e




import pytest

from ansible_bender.conf import Build
from ansible_bender.builders.buildah_builder import BuildahBuilder


@pytest.mark.parametrize("image_name,found", [
    ("registry.fedoraproject.org/fedora:29", True),
    ("docker.io/library/python:3-alpine", True),
    ("docker.io/library/busybox", False),
    ])
def test_find_py_intrprtr_in_fedora_image(image_name, found):
    build = Build()
    build.base_image = image_name
    build.target_image = "starena"
    bb = BuildahBuilder(build)
    try:
        assert bb.find_python_interpreter()
    except RuntimeError:
        if found:
            # interpreter should have been found
            raise



import json

from ansible_bender.builders import buildah_builder
from ansible_bender.builders.buildah_builder import get_buildah_image_id
from flexmock import flexmock

from tests.spellbook import buildah_inspect_data_path


def mock_inspect():
    with open(buildah_inspect_data_path) as fd:
        buildah_inspect_data = json.load(fd)
    flexmock(buildah_builder, inspect_resource=lambda x, y: buildah_inspect_data)


def test_buildah_id():
    mock_inspect()
    assert get_buildah_image_id("this-is-mocked") == "6aed6d59a707a7040ad25063eafd3a2165961a2c9f4d1d06ed0a73bdf2a89322"




OUT_LOGGER = "ab-out"
OUT_LOGGER_FORMAT = "%(message)s"
TIMESTAMP_FORMAT = "%Y%m%d-%H%M%S%f"
TIMESTAMP_FORMAT_TOGETHER = "%Y%m%d%H%M%S%f"
NO_CACHE_TAG = "no-cache"

# configuration related constants
ANNOTATIONS_KEY = "annotations"



import pytest

from ansible_bender.cli import split_once_or_fail_with


def test_split_once():
    secret = "banana"
    with pytest.raises(RuntimeError, match=secret):
        split_once_or_fail_with("a-a-a", "=", secret)
    assert ("a", "a") == split_once_or_fail_with("a=a", "=", secret)



import json
import logging
import subprocess

from ansible_bender.builders.base import Builder
from ansible_bender.utils import graceful_get, run_cmd, buildah_command_exists, \
    podman_command_exists

logger = logging.getLogger(__name__)


def inspect_resource(resource_type, resource_id):
    try:
        i = run_cmd(["buildah", "inspect", "-t", resource_type, resource_id],
                    return_output=True, log_output=False)
    except subprocess.CalledProcessError:
        logger.info("no such %s %s", resource_type, resource_id)
        return None
    try:
        metadata = json.loads(i)
    except IndexError:
        logger.info("no such %s %s", resource_type, resource_id)
        return None
    return metadata


def get_buildah_image_id(container_image):
    metadata = inspect_resource("image", container_image)
    return graceful_get(metadata, "FromImageID")


def pull_buildah_image(container_image):
    run_cmd(["buildah", "pull", container_image],
            save_output_in_exc=False,
            log_stderr=False, print_output=True, log_output=False)


def does_image_exist(container_image):
    run_cmd(["podman", "image", "exists", container_image])


def podman_run_cmd(container_image, cmd, log_stderr=True, return_output=False):
    """
    run provided command in selected container image using podman; raise exc when command fails

    :param container_image: str
    :param cmd: list of str
    :param log_stderr: bool, log errors to stdout as ERROR level
    :param return_output: bool, if True, return output of the command
    :return: stdout output
    """
    return run_cmd(["podman", "run", "--rm", container_image] + cmd,
                   return_output=return_output, log_stderr=log_stderr)


def create_buildah_container(container_image, container_name, build_volumes=None, debug=False):
    """
    Create new buildah container according to spec.

    :param container_image: name of the image
    :param container_name: name of the container to work in
    :param build_volumes: list of str, bind-mount specification: ["/host:/cont", ...]
    :param debug: bool, make buildah print debug info
    """
    args = []
    if build_volumes:
        args += ["-v"] + build_volumes
    args += ["--name", container_name, container_image]
    # will pull the image by default if it's not present in buildah's storage
    buildah("from", args, debug=debug, log_stderr=True)


def configure_buildah_container(container_name, working_dir=None, env_vars=None,
                                labels=None, annotations=None,
                                user=None, cmd=None, ports=None, volumes=None,
                                debug=False):
    """
    apply metadata on the container so they get inherited in an image

    :param container_name: name of the container to work in
    :param working_dir: str, path to a working directory within container image
    :param labels: dict with labels
    :param annotations: dict with annotations
    :param env_vars: dict with env vars
    :param cmd: str, command to run by default in the container
    :param user: str, username or uid; the container gets invoked with this user by default
    :param ports: list of str, ports to expose from container by default
    :param volumes: list of str; paths within the container which has data stored outside
                    of the container
    :param debug: bool, make buildah print debug info
    """
    config_args = []
    if working_dir:
        config_args += ["--workingdir", working_dir]
    if env_vars:
        for k, v in env_vars.items():
            config_args += ["-e", "%s=%s" % (k, v)]
    if labels:
        for k, v in labels.items():
            config_args += ["-l", "%s=%s" % (k, v)]
    if annotations:
        for k, v in annotations.items():
            config_args += ["--annotation", "%s=%s" % (k, v)]
    if user:
        config_args += ["--user", user]
    if cmd:
        config_args += ["--cmd", cmd]
    if ports:
        for p in ports:
            config_args += ["-p", p]
    if volumes:
        for v in volumes:
            config_args += ["-v", v]
    if config_args:
        buildah("config", config_args + [container_name], debug=debug)
    return container_name


def buildah(command, args_and_opts, print_output=False, debug=False, log_stderr=False):
    cmd = ["buildah"]
    # if debug:
    #     cmd += ["--debug"]
    cmd += [command] + args_and_opts
    logger.debug("running command: %s", command)
    return run_cmd(cmd, print_output=print_output, log_stderr=log_stderr)


def buildah_with_output(command, args_and_opts, debug=False):
    cmd = ["buildah"]
    # if debug:
    #     cmd += ["--debug"]
    cmd += [command] + args_and_opts
    output = run_cmd(cmd, return_output=True)
    logger.debug("output: %s", output)
    return output


class BuildahBuilder(Builder):
    ansible_connection = "buildah"
    name = "buildah"

    def __init__(self, build, debug=False):
        """
        :param build: instance of Build
        :param debug: bool, run buildah in debug or not?
        """
        super().__init__(build, debug=debug)
        self.target_image = build.target_image
        self.ansible_host = build.build_container
        self.logs = []
        buildah_command_exists()
        podman_command_exists()

    def create(self):
        """
        create a container where all the work happens
        """
        create_buildah_container(
            self.build.get_top_layer_id(), self.ansible_host,
            build_volumes=self.build.build_volumes, debug=self.debug)
        # let's apply configuration before execing the playbook, except for user
        configure_buildah_container(
            self.ansible_host, working_dir=self.build.metadata.working_dir,
            env_vars=self.build.metadata.env_vars,
            ports=self.build.metadata.ports,
            labels=self.build.metadata.labels,  # labels are not applied when they are configured
                                                # before doing commit
            annotations=self.build.metadata.annotations,
            debug=self.debug
        )

    def run(self, image_name, command):
        """
        run provided command in the selected image and return output

        :param image_name: str
        :param command: list of str
        :return: str (output)
        """
        cmd = ["podman", "run", "--rm", image_name] + command
        return run_cmd(cmd, return_output=True)

    def swap_working_container(self):
        """
        remove current working container and replace it with the provided one
        """
        self.clean()
        self.create()

    def commit(self, image_name, print_output=True):
        if self.build.metadata.user or self.build.metadata.cmd or self.build.metadata.volumes:
            # change user if needed
            configure_buildah_container(
                self.ansible_host, user=self.build.metadata.user,
                cmd=self.build.metadata.cmd,
                volumes=self.build.metadata.volumes,
            )
        buildah("commit", [self.ansible_host, image_name], print_output=print_output,
                debug=self.debug)
        return self.get_image_id(image_name)

    def clean(self):
        """
        clean working container
        """
        buildah("rm", [self.ansible_host], debug=self.debug)

    def get_image_id(self, image_name):
        """ return image_id for provided image """
        image_id = get_buildah_image_id(image_name)
        if not image_id:
            raise RuntimeError("We haven't got any image ID: the image is not present "
                               "or buildah/podman is malfunctioning.")
        return image_id

    def is_image_present(self, image_reference):
        """
        :return: True when the selected image is present, False otherwise
        """
        if not image_reference:
            return False
        try:
            does_image_exist(image_reference)
        except subprocess.CalledProcessError:
            return False
        else:
            return True

    def pull(self):
        """
        pull base image
        """
        logger.info("pulling base image: %s", self.build.base_image)
        pull_buildah_image(self.build.base_image)

    def push(self, build, target, force=False):
        """
        push built image into a remote location using `podman push`

        :param target: str, transport:details
        :param build: instance of Build
        :param force: bool, bypass checks if True
        :return: None
        """
        built_image = build.get_target_image_id()
        cmd = ["buildah", "push", built_image, target]
        # podman prints progress to stderr
        run_cmd(cmd, print_output=False, log_stderr=False)

    def find_python_interpreter(self):
        """
        find python executable in the base image

        :return: str, path to python interpreter
        """
        for i in self.python_interpr_prio:
            cmd = ["ls", i]
            try:
                run_cmd(["podman", "run", "--rm", self.build.base_image] + cmd,
                        log_stderr=False, log_output=True)
            except subprocess.CalledProcessError:
                logger.info("python interpreter %s does not exist", i)
                continue
            else:
                logger.info("using python interpreter %s", i)
                return i
        logger.error("couldn't locate python interpreter, tried these paths: %s", self.python_interpr_prio)
        raise RuntimeError(f"no python interpreter was found in the base image \"{self.build.base_image}\""
                           ", you can specify the path via CLI option --python-interpreter")

    def get_logs(self):
        """
        obtain logs for the selected build

        :return: list of str
        """

    def sanity_check(self):
        """
        invoke container tooling and thus verify they work well
        """
        # doing podman info would be super-handy, but it will immensely clutter the output
        logger.debug("checking that podman command works")
        run_cmd(["podman", "version"], log_stderr=True, log_output=True)
        logger.debug("checking that buildah command works")
        run_cmd(["buildah", "version"], log_stderr=True, log_output=True)

    def check_container_creation(self):
        """
        check that containers can be created
        """
        logger.debug("trying to create a dummy container using podman")
        podman_run_cmd(self.build.base_image, ["true"], log_stderr=True)



import importlib
from functools import partial
from pathlib import Path

import pytest
from flexmock import flexmock

from ansible_bender.core import PbVarsParser


def mock_read_text(return_val=None, raise_exc=False):
    if raise_exc:
        def _f():
            raise FileNotFoundError()
        flexmock(Path, read_text=_f)
    else:
        flexmock(Path, read_text=lambda: return_val)


def mock_import_module(raise_exc=False):
    if raise_exc:
        def _f(name, package=None):
            raise ModuleNotFoundError()
        flexmock(importlib, import_module=_f)
    else:
        flexmock(importlib, import_module=lambda name: None)


@pytest.mark.parametrize("mock_r_t,mock_i_m,should_raise", (
    (
        partial(mock_read_text, "1"),
        partial(mock_import_module, False),
        False
    ),
    (
        partial(mock_read_text, "1"),
        partial(mock_import_module, True),
        True
    ),
    (
        partial(mock_read_text, "0"),
        partial(mock_import_module, False),
        False
    ),
    (
        partial(mock_read_text, "0"),
        partial(mock_import_module, True),
        True
    ),
    (
        partial(mock_read_text, None, True),
        partial(mock_import_module, False),
        False
    ),
))
def test_ansible_selinux_workaround(mock_r_t, mock_i_m, should_raise):
    mock_r_t()
    mock_i_m()
    p = PbVarsParser("")
    if should_raise:
        with pytest.raises(RuntimeError) as ex:
            p._check_selinux_iz_gud()
        assert "libselinux" in str(ex.value)
    else:
        p._check_selinux_iz_gud()



"""
Module to interact with Ansible, perform ansible-pleybook and extract metadata from Ansible vars

A sample configuration:

- hosts: all
  vars:
    ansible_bender:

      base_image: fedora:28

      target_image:
        name: asdqwe
        # configure environment variables: same name as ansible
        environment:
          X: Y
        labels:
          key: value
        # default working directory
        working_dir: /path

      working_container:
        user: 12345  # TODO
        volumes:
        - { src: /path/to/my/code, dest: /src:Z }

tasks:
  ...

"""
import copy
import datetime
import importlib
import json
import logging
import os
import shlex
import shutil
import subprocess
import sys
import tempfile
from pathlib import Path

import yaml

import ansible_bender
from ansible_bender import callback_plugins
from ansible_bender.conf import ImageMetadata, Build
from ansible_bender.constants import TIMESTAMP_FORMAT, TIMESTAMP_FORMAT_TOGETHER
from ansible_bender.exceptions import AbBuildUnsuccesful
from ansible_bender.utils import run_cmd, ap_command_exists, random_str, graceful_get, \
    is_ansibles_python_2

logger = logging.getLogger(__name__)
A_CFG_TEMPLATE = """\
[defaults]
# when user is changed, ansible might not be able to write to /.ansible
remote_tmp = /tmp
callback_plugins={0}
callback_whitelist=snapshoter\n
"""


def run_playbook(playbook_path, inventory_path, a_cfg_path, connection, extra_variables=None,
                 ansible_args=None, debug=False, environment=None, try_unshare=True,
                 provide_output=True, log_stderr=False):
    """
    run selected ansible playbook and return output from ansible-playbook run

    :param playbook_path:
    :param inventory_path:
    :param a_cfg_path:
    :param connection:
    :param extra_variables:
    :param ansible_args: list of str, extra arguments for a-p
    :param debug:
    :param environment:
    :param try_unshare: bool, do `buildah unshare` if uid > 0
    :param provide_output: bool, present output to user
    :param log_stderr: bool, log errors coming from stderr to our logger

    :return: output
    """
    ap = ap_command_exists()
    if is_ansibles_python_2(ap):
        # I just realized it could work, we would just had to disable the
        # callback plugin: no caching and layering
        raise RuntimeError(
            "ansible-bender is written in python 3 and does not work in python 2,\n"
            f"it seems that {ap} is using python 2 - ansible-bender will not"
            "work in such environment\n"
        )
    cmd_args = [
        ap,
        "-c", connection,

    ]
    if inventory_path:
        cmd_args += ["-i", inventory_path]
    if debug:
        cmd_args += ["-vvv"]
    if extra_variables:
        cmd_args += ["--extra-vars"] + \
                    [" ".join(
                        ["{}={}".format(k, v)
                         for k, v in extra_variables.items()]
                    )]
    if ansible_args:
        cmd_args += ansible_args
    cmd_args += [playbook_path]
    logger.debug("%s", " ".join(cmd_args))

    env = os.environ.copy()
    env["ANSIBLE_RETRY_FILES_ENABLED"] = "0"
    if environment:
        env.update(environment)
    if a_cfg_path:
        env["ANSIBLE_CONFIG"] = a_cfg_path

    if try_unshare and os.getuid() != 0:
        logger.info("we are running rootless, prepending `buildah unshare`")
        # rootless, we need to `buildah unshare` for sake of `buildah mount`
        # https://github.com/containers/buildah/issues/1271
        # the need for `--` https://github.com/containers/buildah/issues/1374
        cmd_args = ["buildah", "unshare", "--"] + cmd_args

    # ansible has no official python API, the API they have is internal and said to break compat
    try:
        return run_cmd(
            cmd_args,
            print_output=provide_output,
            save_output_in_exc=True,
            env=env,
            return_all_output=provide_output,
            log_stderr=log_stderr,
        )
    except subprocess.CalledProcessError as ex:
        raise AbBuildUnsuccesful("ansible-playbook execution failed: %s" % ex, ex.output)


class AnsibleRunner:
    """
    Run ansible on provided artifact using the Builder interface
    """

    def __init__(self, playbook_path, builder, build, debug=False):
        self.build_i = build
        self.pb = playbook_path
        self.builder = builder
        self.debug = debug

    def _create_inventory_file(self, fd, python_interpreter):
        fd.write(
            '%s ansible_connection="%s" ansible_python_interpreter="%s"\n' % (
                self.builder.ansible_host,
                self.builder.ansible_connection,
                python_interpreter
            )
        )

    def _create_ansible_cfg(self, fd):
        callback_plugins_dir = os.path.dirname(callback_plugins.__file__)
        fd.write(A_CFG_TEMPLATE.format(callback_plugins_dir))

    def _get_path_our_site(self):
        """ return a path to a directory which contains ansible_bender installation """
        # pip in Fedora installs to /usr/local which is on default pythonpath
        # but when ansible invokes the callback plugin, that directory is not on sys.path: wat?
        # hence, let's add the site ab is installed in to sys.path
        return os.path.dirname(os.path.dirname(ansible_bender.__file__))

    def build(self, db_path):
        """
        run the playbook against the container

        :param db_path, str, path to ab's database

        :return: str, output
        """
        tmp = tempfile.mkdtemp(prefix="ab")

        pythonpath = os.environ.get("PYTHONPATH", "")
        site_path = self._get_path_our_site()
        if pythonpath:
            pythonpath = f"{site_path}:{pythonpath}"
        else:
            pythonpath = site_path

        try:
            environment = {
                "AB_BUILD_ID": self.build_i.build_id,
                "AB_DB_PATH": db_path,
                "PYTHONPATH": pythonpath,  # TODO write an e2e test for this
            }
            inv_path = os.path.join(tmp, "inventory")
            logger.info("creating inventory file %s", inv_path)
            with open(inv_path, "w") as fd:
                self._create_inventory_file(fd, self.build_i.python_interpreter)
            a_cfg_path = os.path.join(tmp, "ansible.cfg")
            with open(a_cfg_path, "w") as fd:
                self._create_ansible_cfg(fd)

            tmp_pb_path = os.path.join(tmp, "p.yaml")
            with open(self.pb, "r") as fd_r:
                pb_dict = yaml.safe_load(fd_r)
            for idx, doc in enumerate(pb_dict):
                host = doc["hosts"]
                logger.debug("play[%s], host = %s", idx, host)
                doc["hosts"] = self.builder.ansible_host
            with open(tmp_pb_path, "w") as fd:
                yaml.safe_dump(pb_dict, fd)
            playbook_base = os.path.basename(self.pb).split(".", 1)[0]
            timestamp = datetime.datetime.now().strftime(TIMESTAMP_FORMAT)
            symlink_name = f".{playbook_base}-{timestamp}-{random_str()}.yaml"
            playbook_dir = os.path.dirname(self.pb)
            symlink_path = os.path.join(playbook_dir, symlink_name)
            os.symlink(tmp_pb_path, symlink_path)

            extra_args = None
            try:
                if self.build_i.ansible_extra_args:
                    extra_args = shlex.split(self.build_i.ansible_extra_args)
                return run_playbook(
                    symlink_path, inv_path, a_cfg_path, self.builder.ansible_connection,
                    debug=self.debug, environment=environment, ansible_args=extra_args
                )
            finally:
                os.unlink(symlink_path)
        finally:
            shutil.rmtree(tmp)


class PbVarsParser:
    def __init__(self, playbook_path):
        """
        :param playbook_path: str, path to playbook
        """
        self.playbook_path = playbook_path
        self.build = Build()
        self.metadata = ImageMetadata()
        self.build.metadata = self.metadata

    def _check_selinux_iz_gud(self):
        """
        This is a workaround for a weird behavior of ansible: if selinux is
        in the enforcing mode and python3-libselinux is not installed, ansible freezes

        https://bugzilla.redhat.com/show_bug.cgi?id=1696706
        :return:
        """
        try:
            enforcing_status = Path("/sys/fs/selinux/enforce").read_text()
        except FileNotFoundError:
            logger.debug("this system is not using selinux, /sys/fs/selinux/enforce is not present")
            return
        logger.debug(f"selinux enforce status = {enforcing_status}")
        # it can be enforcing or not, selinux python module needs to be present
        try:
            importlib.import_module("selinux")
        except ModuleNotFoundError:
            raise RuntimeError(
                "\nThis system is using selinux(8) and selinux python module is not installed. "
                "There is a known issue in ansible that it freezes in this setup:\n"
                "  https://bugzilla.redhat.com/show_bug.cgi?id=1696706\n"
                "Please install libselinux python bindings (on Fedora the package name is python3-libselinux)."
            )

    def expand_pb_vars(self):
        """
        populate vars from a playbook, defined in vars section

        :return: dict with the content of ansible_bender var
        """
        self._check_selinux_iz_gud()
        with open(self.playbook_path) as fd:
            plays = yaml.safe_load(fd)

        for play in plays[1:]:
            bender_vars = graceful_get(play, "vars", "ansible_bender")
            if bender_vars:
                logger.warning("Variables are loaded only from the first play.")

        try:
            # we care only about the first play, we don't want to merge dicts
            d = plays[0]
        except IndexError:
            raise RuntimeError("Invalid playbook, can't access the first document.")

        bender_vars = graceful_get(d, "vars", "ansible_bender")
        if not bender_vars:
            logger.info("no bender data found in the playbook")
            return {}

        tmp = tempfile.mkdtemp(prefix="ab")
        json_data_path = os.path.join(tmp, "j.json")

        # we cannot use "vars" variable because the variables are not expanded in there
        pb_vars = copy.deepcopy(d["vars"])
        while True:
            # just in case the variable is already defined
            timestamp = datetime.datetime.now().strftime(TIMESTAMP_FORMAT_TOGETHER)
            ab_vars_key = f"ab_vars_{timestamp}"
            if ab_vars_key not in pb_vars:
                logger.debug("ab vars key = %s", ab_vars_key)
                pb_vars[ab_vars_key] = d["vars"]
                break
        jinja_pb_vars_key = '{{ %s }}' % ab_vars_key
        pb = {
            "name": "Let Ansible expand variables",
            "hosts": "localhost",
            "vars": pb_vars,
            "vars_files": d.get("vars_files", []),
            "gather_facts": False,
            "tasks": [
                {"debug": {"msg": jinja_pb_vars_key}},
                {
                    "copy": {
                        "dest": json_data_path,
                        "content": jinja_pb_vars_key
                    }
                }
            ]
        }

        i_path = os.path.join(tmp, "i")
        with open(i_path, "w") as fd:
            fd.write("localhost ansible_connection=local")

        # json is easier to parse than yaml
        tmp_pb_path = os.path.join(tmp, "p.json")
        with open(tmp_pb_path, "w") as fd:
            json.dump([pb], fd)

        playbook_base = os.path.basename(self.playbook_path).split(".", 1)[0]
        symlink_name = f".{playbook_base}-{timestamp}-{random_str()}.yaml"
        playbook_dir = os.path.dirname(self.playbook_path)
        symlink_path = os.path.join(playbook_dir, symlink_name)
        os.symlink(tmp_pb_path, symlink_path)

        # yeah, ansible is not very smart for connection=local
        args = ["-e", f"ansible_python_interpreter={sys.executable}"]

        try:
            run_playbook(symlink_path, i_path, None, connection="local", try_unshare=False,
                         provide_output=False, log_stderr=True, ansible_args=args)

            with open(json_data_path) as fd:
                return json.load(fd)["ansible_bender"]
        finally:
            os.unlink(symlink_path)
            shutil.rmtree(tmp)

    def process_pb_vars(self, bender_data):
        """
        accept variables from the playbook and update the Build and ImageMetadata objects with them

        :param bender_data: dict with the content of ansible_bender vars
        :return:
        """
        if not bender_data:
            return
        self.metadata.update_from_configuration(bender_data.get("target_image", {}))
        self.build.update_from_configuration(bender_data)

    def get_build_and_metadata(self):
        """
        extra vars from the selected playbook

        :return: Build(), ImageMetadata()
        """
        bender_data = self.expand_pb_vars()

        self.process_pb_vars(bender_data)

        return self.build, self.metadata



import os

import jsonschema
import pytest
from ansible_bender.utils import set_logging

from ansible_bender.conf import ImageMetadata, Build
from ansible_bender.core import PbVarsParser

from tests.spellbook import b_p_w_vars_path, basic_playbook_path, full_conf_pb_path, multiplay_path


def test_expand_pb_vars():
    p = PbVarsParser(b_p_w_vars_path)
    data = p.expand_pb_vars()
    assert data["base_image"] == "docker.io/library/python:3-alpine"
    assert data["ansible_extra_args"] == "-vvv"
    playbook_dir = os.path.dirname(b_p_w_vars_path)
    assert data["working_container"]["volumes"] == [f"{playbook_dir}:/src:Z"]
    assert data["target_image"]["name"] == "challet"
    assert data["target_image"]["labels"] == {"x": "y"}
    assert data["target_image"]["environment"] == {"asd": playbook_dir}


def test_b_m_empty():
    """ test that build and metadata are 'empty' when there are no vars """
    p = PbVarsParser(basic_playbook_path)
    b, m = p.get_build_and_metadata()
    b.playbook_path = "/somewhere.yaml"

    b.base_image = "fedora:29"
    b.playbook_path = "/asd.yaml"
    b.target_image = "lolz"

    b.validate()
    m.validate()
    assert isinstance(b, Build)
    assert isinstance(m, ImageMetadata)
    assert b.cache_tasks is True
    assert b.layering is True


def test_set_all_params():
    """ test that we can set all the parameters """
    p = PbVarsParser(full_conf_pb_path)
    b, m = p.get_build_and_metadata()
    b.playbook_path = "/somewhere.yaml"

    b.validate()
    m.validate()

    assert isinstance(b, Build)
    assert isinstance(m, ImageMetadata)

    assert b.base_image == "mona_lisa"
    assert b.layering
    assert not b.cache_tasks
    assert b.ansible_extra_args == "--some --args"
    assert b.build_volumes == ["/c:/d"]
    assert b.target_image == "funky-mona-lisa"

    assert m.env_vars == {"z": "value"}
    assert m.volumes == ["/a"]
    assert m.working_dir == "/workshop"
    assert m.labels == {"x": "y"}
    assert m.annotations == {"bohemian": "rhapsody"}
    assert m.cmd == "command -x -y z"
    assert m.user == "leonardo"


def test_validation_err_ux():
    """ Test that validation errors are useful """
    p = PbVarsParser(basic_playbook_path)
    b, m = p.get_build_and_metadata()

    with pytest.raises(jsonschema.exceptions.ValidationError) as ex:
        b.validate()

    s = str(ex.value)

    assert "is not of type" in s
    assert "Failed validating 'type' in schema" in s


def test_multiplay(caplog):
    set_logging()
    p = PbVarsParser(multiplay_path)
    b, m = p.get_build_and_metadata()

    assert b.target_image != "nope"
    assert "Variables are loaded only from the first play." == caplog.records[0].msg
    assert "no bender data found in the playbook" == caplog.records[1].msg



""" Making sure that ab can be used as a custom builder in okd """

import json
import os

from flexmock import flexmock

from ansible_bender.api import Application
from ansible_bender.okd import build_inside_openshift


# OKD sets an env var BUILD to this VALUE
# oc get --template '{{ (index (index .spec.containers 0).env 0).value }}' pod/ab-in-okd-1-build | jq
# fun fun
BUILD_ENV = {
  "kind": "Build",
  "apiVersion": "build.openshift.io/v1",
  "metadata": {
    "name": "ab-in-okd-1",
    "namespace": "myproject",
    "selfLink": "/apis/build.openshift.io/v1/namespaces/myproject/builds/ab-in-okd-1",
    "uid": "b0f55118-09d8-11e9-8e48-8c164572b096",
    "resourceVersion": "39780",
    "creationTimestamp": "2018-12-27T13:09:49Z",
    "labels": {
      "buildconfig": "ab-in-okd",
      "openshift.io/build-config.name": "ab-in-okd",
      "openshift.io/build.start-policy": "Serial"
    },
    "annotations": {
      "openshift.io/build-config.name": "ab-in-okd",
      "openshift.io/build.number": "1"
    },
    "ownerReferences": [
      {
        "apiVersion": "build.openshift.io/v1",
        "kind": "BuildConfig",
        "name": "ab-in-okd",
        "uid": "afd64f3d-09d8-11e9-8e48-8c164572b096",
        "controller": True
      }
    ]
  },
  "spec": {
    "serviceAccount": "builder",
    "source": {
      "type": "Git",
      "git": {
        "uri": "https://github.com/TomasTomecek/ansible-bender",
        "ref": "master"
      }
    },
    "strategy": {
      "type": "Custom",
      "customStrategy": {
        "from": {
          "kind": "DockerImage",
          "name": "ansible-bender:latest"
        },
        "pullSecret": {
          "name": "builder-dockercfg-mfvxv"
        },
        "env": [
          {
            "name": "AB_BASE_IMAGE",
            "value": "registry.fedoraproject.org/fedora:29"
          },
          {
            "name": "AB_PLAYBOOK_PATH",
            "value": "recipe.yml"
          },
          {
            "name": "OPENSHIFT_CUSTOM_BUILD_BASE_IMAGE",
            "value": "ansible-bender:latest"
          }
        ]
      }
    },
    "output": {
      "to": {
        "kind": "DockerImage",
        "name": "lolzor"
      },
      "pushSecret": {
        "name": "builder-dockercfg-mfvxv"
      }
    },
    "resources": {},
    "postCommit": {},
    "nodeSelector": None,
    "triggeredBy": [
      {
        "message": "Manually triggered"
      }
    ]
  },
  "status": {
    "phase": "New",
    "outputDockerImageReference": "lolzor",
    "config": {
      "kind": "BuildConfig",
      "namespace": "myproject",
      "name": "ab-in-okd"
    },
    "output": {}
  }
}
BUILD_ENV_RAW = json.dumps(BUILD_ENV)


def test_bio(tmpdir):
    database_path = str(tmpdir)
    flexmock(Application, build=lambda build, extra_ansible_args=None: True)
    application = Application(db_path=database_path, debug=True)

    ose = os.environ
    ose["BUILD"] = BUILD_ENV_RAW
    ose["AB_PLAYBOOK_PATH"] = "asdqwe.yml"
    ose["AB_BASE_IMAGE"] = "pancake"
    flexmock(os, environ=ose)
    try:
        build_inside_openshift(application)
    finally:
        application.clean()



"""
Tests for ansible invocation
"""
import pytest
from flexmock import flexmock

from ansible_bender import utils
from ansible_bender.core import run_playbook
from tests.spellbook import C7_AP_VER_OUT


def test_ansibles_python():
    flexmock(utils, run_cmd=lambda *args, **kwargs: C7_AP_VER_OUT),
    with pytest.raises(RuntimeError) as ex:
        run_playbook(None, None, None, None)
    assert str(ex.value).startswith(
        "ansible-bender is written in python 3 and does not work in python 2,\n")
